/* (c) Joerg Schulenburg Jun2000-2007

 this is part of spin-package

 it is for division of matrix multiplication 
 in parallelizable block matrix multiplication (see block.h)

 */
#include "block.h"

int mpi_n=1, mpi_i=0;   /* number of MPI nodes, no. of this node */
int pt_n=1;             /* number of threads per node 1..B_NUM */
int mpi_n_bits=1, pt_n_bits=1; /* bits needed to store mpi_i and pt_i, MPI */
/* ToDo: only 32 bit? -> ULong */
unsigned long node_ofs, node_n1; /* ofset and length of subspace this node */
unsigned long b_ofs[B_NUM+1];	      /* see block.h */
unsigned long mpi_ofs[MPI_MAX_NUM+1]; /* see block.h */
unsigned long max_b_len;        /* maximum b_len to ensure mpi balance */
/* ToDo: move b_hnz to h_hnz h_file.c!? where h_ is defined */
unsigned long b_hnz[B_NUM];	/* block NonZero H-entries */
unsigned long b_snz[B_NUM];	/* stored hnz block entries */
unsigned long b_sn1[B_NUM];	/* stored hnz block lines */
unsigned long len_full_block;   /* number of elements per block */

/* ToDo: move hbuf to h_file.c!? where thxy is defined */
struct thxy b_hbuf[B_NUM];      /* buffer for faster r/w-access */ 
int         b_hbl[B_NUM];       /* num valid data in buffer hbuf */  

/* release mem to satisfy memleak checker */
void b_free(){
   /* free dynamic arrays */
   /* nothing to do v2.36 */
}

/* initialization of thread blocks for n1/mpi_n elements on each node,
 * n1 is whole matrix size, called by every node, only thread0
 */
int b_ini(unsigned long n1,int vvv){
  int i, num_full_blocks;
  unsigned long len_rest_block, ofs, len;
  
  max_b_len=0;
          
  /* pt_n is numthreads per node */
  for (i=0;i<=      B_NUM;i++)   b_ofs[i]=0; /* initialize array 0..pt_n */
  for (i=0;i<=MPI_MAX_NUM;i++) mpi_ofs[i]=0; /* initialize array 0..mpi_n */
  for (i=0;i<B_NUM;i++) {  /* initialize all arrays (1..pt_n..B_NUM)-1 */
    b_sn1[i]=0; /* belongs to h_file!? */
    b_hnz[i]=0;
    b_snz[i]=0;
    b_hbl[i]=0;
  }
  /* block length depends from number of mpi nodes and threads per node */
  len_full_block=(n1+pt_n*mpi_n-1)/(pt_n*mpi_n); /* #loops/thread 0..n1 */
#if defined(HAVE_MMAP) && defined(CFG_PAGESIZE)
  /* mmap offset must be a multiple of PAGESIZE! fill up block length */
  if (len_full_block>0)
      len_full_block=((len_full_block-1)/CFG_PAGESIZE+1)*CFG_PAGESIZE;
#endif /* CFG_PAGESIZE */

  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (!mpi_i)
    printf(" b_ini: mpi_n= %d pt_n= %d n1= %lu len_full_blk= %lu\n",
       mpi_n, pt_n, n1, len_full_block);

  /* calculate num of full size blocks and length rest block */
  num_full_blocks= n1 / len_full_block;     /* range: 0..(pt_n*mpi_n-1) */
  len_rest_block=  n1 - len_full_block * num_full_blocks;  /* 0..(n1-1) */

  if (!mpi_i) {
    printf(" blocks:  num   * length  + rest = %3d * %8lu + %8lu = n1\n",
       num_full_blocks, len_full_block, len_rest_block);
  }

  node_ofs=0; node_n1=0;
  MPI_Barrier(MPI_COMM_WORLD); /* debug */

  /* setup mpi blocks (offsets and implicite lengths) */
  ofs=0; len=n1; max_b_len=len_full_block*pt_n;
  for (i=0; i<mpi_n; i++) {
    mpi_ofs[i  ]=ofs;
    mpi_ofs[i+1]=ofs+((len<max_b_len)?len:max_b_len); /* next-ofs */
    ofs =mpi_ofs[i+1];
    len-=mpi_ofs[i+1]-mpi_ofs[i];
    /* only mpi_i==0 prints out table for other mpi-tasks */
    if (((~vvv)&2) || mpi_i /* || mpi_ofs[i+1]==mpi_ofs[i]*/) continue; /* dont print */
    if ((i==2 && mpi_n>4)) printf("MPI_DBG: ...\n");
    if ((i< 2 || i>=mpi_n-2)) /* ToDo: better if((i&(i-1))==0||i+1==mpi_n) */
      printf("MPI_DBG: node=%3d       mpi_len= %8lu mpi_ofs= %10lu\n",
              i, mpi_ofs[i+1]-mpi_ofs[i], mpi_ofs[i]);
  }
  node_n1 =mpi_ofs[mpi_i+1]
          -mpi_ofs[mpi_i];  /* length of block in this node */
  node_ofs=mpi_ofs[mpi_i];  /* node offset, do we need it? */

  /* setup pthread blocks (offsets and implicite lengths) */
  ofs=0; len=node_n1; max_b_len=len_full_block;
  for (i=0; i<pt_n; i++) {
    b_ofs[i  ]=ofs;
    b_ofs[i+1]=ofs+((len<max_b_len)?len:max_b_len);
    ofs =b_ofs[i+1];
    len-=b_ofs[i+1]-b_ofs[i];
    /* print mpi_i==0 only, mixed code removed later for simplicity? */
    if (((~vvv)&2) || mpi_i || i<pt_n-4 || i>3 || pt_n==1
                  /* || b_ofs[i+1]==b_ofs[i] */) continue; /* dont print */
    printf(" PT_DBG: node=%3d blk=%3d b_len= %8lu   b_ofs= %10lu\n",
              mpi_i, i, b_ofs[i+1]-b_ofs[i], b_ofs[i]);
  }

  return 0;
}
